#include "ShaderConstants.fxh"
#include "util.fxh"

struct PS_Input {
	float4 position : SV_Position;
	float3 screenPos : SPOS;
	float far : FAR;

	#ifndef BYPASS_PIXEL_SHADER
		lpfloat4 color : COLOR;
		snorm float2 uv0 : TEXCOORD_0_FB_MSAA;
		snorm float2 uv1 : TEXCOORD_1_FB_MSAA;
	#endif

	#ifdef NEAR_WATER
		float cameraDist : TEXCOORD_2;
	#endif

	#ifdef FOG
		float4 fogColor : FOG_COLOR;
	#endif
};

struct PS_Output {
	float4 color : SV_Target;
};

//Tone Map
	#define ToneMap
	#define Gamma 1.5

	float Uncharted2Film(float x) {
		float A = 0.22;
		float B = 0.3;
		float C = 0.15 * 1.8;
		float D = 0.4 * 1.3;
		float E = 0.05 * 1.7;
		float F = 0.3;

		return ((x * (A * x + C * B) + D * E) / (x * (A * x + B) + D * F)) - E / F;
	};

	float3 toneMapping(float3 clr) {
		float W = 1.3 / 1.4;

		#ifdef ToneMap
			float Luma = dot(clr, float3(0.0, 0.3, 0.3));
			float3 Chroma = clr - Luma;
			clr = (Chroma * Gamma) + Luma;

			clr = float3(Uncharted2Film(clr.r), Uncharted2Film(clr.g), Uncharted2Film(clr.b)) / Uncharted2Film(W);
		#endif

		return clr;
	};
//End of Tone Map

void main( in PS_Input PSInput, out PS_Output PSOutput ) {
	#ifdef BYPASS_PIXEL_SHADER
	    PSOutput.color = float4(0.0f, 0.0f, 0.0f, 0.0f);
 	   return;
	#else

		#if !defined(TEXEL_AA) || !defined(TEXEL_AA_FEATURE) || (VERSION < 0xa000 /*D3D_FEATURE_LEVEL_10_0*/) 
			float4 diffuse = TEXTURE_0.Sample(TextureSampler0, PSInput.uv0);
		#else
			float4 diffuse = texture2D_AA(TEXTURE_0, TextureSampler0, PSInput.uv0 );
		#endif
		
		float4 blur = float4(0.0);
		blur += TEXTURE_0.Sample(TextureSampler0, PSInput.uv0 + 0.00005);
		blur += TEXTURE_0.Sample(TextureSampler0, PSInput.uv0 + 0.0001);
		blur += TEXTURE_0.Sample(TextureSampler0, PSInput.uv0 + 0.00015);
		blur += TEXTURE_0.Sample(TextureSampler0, PSInput.uv0 + 0.0002);
		blur += TEXTURE_0.Sample(TextureSampler0, PSInput.uv0 + 0.00025);
		blur += TEXTURE_0.Sample(TextureSampler0, PSInput.uv0 + 0.0003);
		blur += TEXTURE_0.Sample(TextureSampler0, PSInput.uv0 + 0.00035);
		blur += TEXTURE_0.Sample(TextureSampler0, PSInput.uv0 + 0.0004);
		blur += TEXTURE_0.Sample(TextureSampler0, PSInput.uv0 + 0.00045);
		blur += TEXTURE_0.Sample(TextureSampler0, PSInput.uv0 + 0.0005);
		blur /= 9.5;

		float dof_factor = length(PSInput.screenPos.xy*1.2);
		diffuse.rgb = lerp( diffuse.rgb, blur.rgb, dof_factor );

		//Torch
			//Main
			float3 torchLight = float3(2.0, 0.4, 0.2);

			//Output
			diffuse.rgb += torchLight * max(dot(0.5, PSInput.uv1.r - 0.5), 0.0);

		//Shadow
			//Main
			float worldShadow = 0.0;
			float nolight = (1.0 - PSInput.uv1.r);

			//Shadow Function
			float SBase = 0.900;
			float SLine = 5.0;

			//Color
			float3 shadowColor = float3(0.0, 0.0, 0.0);
			float toneSat = 1.0;

			//Function
			#ifdef FANCY //Advance Shadow
				if(PSInput.uv1.g < (SBase - 0.001 * SLine)) {worldShadow = 0.05;}
				if(PSInput.uv1.g < (SBase - 0.002 * SLine)) {worldShadow = 0.10;}
				if(PSInput.uv1.g < (SBase - 0.003 * SLine)) {worldShadow = 0.15;}
				if(PSInput.uv1.g < (SBase - 0.004 * SLine)) {worldShadow = 0.20;}
				if(PSInput.uv1.g < (SBase - 0.005 * SLine)) {worldShadow = 0.25;}
				if(PSInput.uv1.g < (SBase - 0.006 * SLine)) {worldShadow = 0.30;}
				if(PSInput.uv1.g < (SBase - 0.007 * SLine)) {worldShadow = 0.35;}
				if(PSInput.uv1.g < (SBase - 0.008 * SLine)) {worldShadow = 0.40;}
				if(PSInput.uv1.g < (SBase - 0.009 * SLine)) {worldShadow = 0.45;}
			#else //Basic Shadow
				if(PSInput.uv1.g < 0.9) {worldShadow = 0.15;}
			#endif

			//Output
			diffuse.rgb += lerp(diffuse.rgb, min(shadowColor, toneSat), worldShadow * nolight);

		#ifdef SEASONS_FAR
			diffuse.a = 1.0f;
			PSInput.color.b = 1.0f;
		#endif

		#ifdef ALPHA_TEST
		//If we know that all the verts in a triangle will have the same alpha, we should cull there first.
			#ifdef ALPHA_TO_COVERAGE
				float alphaThreshold = .05f;
			#else
				float alphaThreshold = .5f;
			#endif

			if(diffuse.a < alphaThreshold)
				discard;
		#endif

		#if !defined(ALWAYS_LIT)
			diffuse = diffuse * TEXTURE_1.Sample(TextureSampler1, PSInput.uv1);
		#endif

		#ifndef SEASONS

		#if !defined(ALPHA_TEST) && !defined(BLEND)
			diffuse.a = PSInput.color.a;
		#elif defined(BLEND)
			#ifdef NEAR_WATER	
				diffuse.a *= PSInput.color.a;

				float alphaFadeOut = saturate(PSInput.cameraDist.x);
				diffuse.a = lerp(diffuse.a, 1.0f, alphaFadeOut);
			#endif
		#endif

		diffuse.rgb *= PSInput.color.rgb;

		#else
			float2 uv = PSInput.color.xy;
			diffuse.rgb *= lerp(1.0f, TEXTURE_2.Sample(TextureSampler2, uv).rgb*2.0f, PSInput.color.b);
			diffuse.rgb *= PSInput.color.aaa;
			diffuse.a = 1.0f;
		#endif

		#ifdef FOG
			diffuse.rgb = lerp( diffuse.rgb, PSInput.fogColor.rgb, PSInput.fogColor.a );
		#endif
		diffuse.rgb = lerp(diffuse.rgb, FOG_COLOR.rgb, PSInput.far);

		diffuse.rgb = toneMapping(diffuse.rgb);
		PSOutput.color = diffuse;

		#ifdef VR_MODE
			// On Rift, the transition from 0 brightness to the lowest 8 bit value is abrupt, so clamp to 
			// the lowest 8 bit value.
			PSOutput.color = max(PSOutput.color, 1 / 255.0f);
		#endif

	#endif // BYPASS_PIXEL_SHADER
}